(************************************************************************)
(*                                                                      *)
(*                    Version 1.1a without error checks                 *)
(*                                                                      *)
(************************************************************************)



IMPLEMENTATION MODULE LISTs;

(************************************************************************)
(*                                                                      *)
(*                              L I S T s                               *)
(*                                                                      *)
(************************************************************************)
(*                                                                      *)
(*           Version 1.1a for Sun/4, SunOS 4.1.1, m2c-Compiler          *)
(*     Version 1.1b for IBM-PC/AT, MSDOS 3.30, Taylor Modula-2 3.00     *)
(*     Version 1.1c for IBM-PC/AT, MSDOS 3.30, Fitted Modula-2 2.00     *)
(*                                                                      *)
(*                    (c) Richard Breuer 1990, 1991                     *)
(*            e-mail: ricki@rwthi3.informatik.rwth-aachen.de            *)
(*              mail: Brunssumstr. 6, 5100 Aachen, Germany              *)
(*                                                                      *)
(************************************************************************)
(*                                                                      *)
(* Notes:                                                               *)
(*                                                                      *)
(*   LISTs is an  implementation of the concept of double linked lists.	*)
(*   The lists are kept in main memory.  Thus the number of elements is	*)
(*   limited only by the computers heap. I call this genericity.	*)
(*									*)
(*   Genericity is implemented by using the generic Modula-2 data types *)
(*   ARRAY OF BYTE  and  ADDRESS.  There is only  one instance of LISTs *)
(*   during  the runtime of an  application no matter how many lists of	*)
(*   various kinds will be defined. The major disadvantage of the usage *)
(*   of the generic  data types is  that the  compiler has no chance to	*)
(*   check for data  type consistency at compile time.  So the applica-	*)
(*   tion must be careful not to affect unwanted side effects.          *)
(*                                                                      *)
(*   Each  list  has two  dummy nodes  Top  and  Bottom  for efficiency	*)
(*   reasons.  They contain no data and help to keep the  pointer algo-	*)
(*   rithms simple.							*)
(*									*)
(************************************************************************)
(*                                                                      *)
(* Revisions:                                                         	*)
(*                                                                      *)
(*   08/21/91	LISTInsertListXXXCursor: Forgot the deletion of top and	*)
(*		bottom node of the second list [BR]			*)
(*   07/24/91	Error in LISTGotoPrevious/Next corrected: There was a	*)
(*		difference in the function of procedures and the	*)
(* 		descriptions in the interface [BR]			*)
(*   03/10/91	Version 1.1c [BR]					*)
(*   02/27/91	Versions 1.1a,1.1b [BR]					*)
(*   02/25/91	LISTExchangePos: No error, if Pos1=Pos2. Just a return	*)
(*		without any action [BR]					*)
(*   02/24/91	Inserted error checks for Allocates [BR]		*)
(*   10/22/90	LISTReplaceCurrent: Error check for undefined cursor	*)
(*		[BR]							*)
(*   09/27/90	Error check whether the given data elements are long	*)
(*		enough for the class [BR]				*)
(*   09/27/90	New procedure LISTCyclicStep [BR]			*)
(*   09/17/90	Versions 1.0a,1.0b [BR]					*)
(*                                                                      *)
(************************************************************************)

  FROM SYSTEM IMPORT
    ADDRESS,
    ADR,
    CCALL,
    TSIZE,
    BYTE;



  FROM Storage IMPORT
    ALLOCATE,
    DEALLOCATE;


  CONST
    MaxClasses   = 256;         
				(* maximal number of list classes to be	*)
				(* defined				*)

  TYPE
    NodeTypePtr = POINTER TO NodeType;
				(* pointer to a list node		*)

    NodeType = RECORD           
				(* record with a list node description	*)
      Prev   : NodeTypePtr;     
				(* - Predecessor			*)
      Next   : NodeTypePtr;     
				(* - Successor				*)
      DataPtr: ADDRESS
                                (* - Pointer to data			*)
    END;

    LISTType = POINTER TO ListType;
				(* generic list type			*)

    ListType = RECORD
			        (* record for list information		*)
      Top    : NodeTypePtr;     
				(* - pointer to top node		*)
      Bottom : NodeTypePtr;     
				(* - pointer to bottom node		*)
      Cursor : NodeTypePtr;     
				(* - pointer to current node		*)
      N      : CARDINAL;        
				(* - number of elements			*)
      CurPos : CARDINAL;        
				(* - current position			*)
      ClassID: CARDINAL         
				(* - class of the list			*)
    END;

    ClassType = RECORD          
    				(* record with class information	*)
      Cmp  : LISTCompareProcType;
                                (* - compare procedure			*)
      Bytes: CARDINAL           
      				(* - element length in BYTEs		*)
    END;

  VAR
    Classes: ARRAY [0..MaxClasses-1] OF ClassType;
				(* array which contains the class	*)
				(* descriptions				*)

    ClassN : CARDINAL;          
				(* number of classes currently defined	*)


  (**********************************************************************)
  (*                                                                    *)
  (*			      Local Procedures				*)
  (*                                                                    *)
  (**********************************************************************)


  PROCEDURE Copy(    From: ADDRESS;
                     To  : ADDRESS;
                     Len : CARDINAL);
  (* Copies Len bytes from From to To. A 'clean' Modula-2 solution does not
     seem to be possible. Please tell me, if you find one! *)

  BEGIN
    (* Aufruf der C-Prozedur 'bcopy' *)
    CCALL("bcopy",From,To,Len)
  END Copy;

  PROCEDURE DeleteCurrent(VAR List: LISTType);
  (* Deletes the current list element an updates the current position. The
     cursor must be defined *)

  VAR
    Help: NodeTypePtr;

  BEGIN
    (* Deallocate the data of the current element *)
    DEALLOCATE(List^.Cursor^.DataPtr,Classes[List^.ClassID].Bytes);
    List^.Cursor^.Prev^.Next := List^.Cursor^.Next;
    List^.Cursor^.Next^.Prev := List^.Cursor^.Prev;
    (* Determine the new current position *)
    Help := List^.Cursor^.Next;
    IF (Help=List^.Bottom) THEN
      Help := List^.Cursor^.Prev;
      (* No proedecessor --> cursor undefined *)
      IF (Help=List^.Top) THEN List^.CurPos := 0
                          ELSE DEC(List^.CurPos)
      END
    END;
    (* Deallocate node *)
    DEALLOCATE(List^.Cursor,TSIZE(NodeType));
    (* Update organisatoric values *)
    DEC(List^.N);
    List^.Cursor := Help
  END DeleteCurrent;

  PROCEDURE GotoElement(VAR List : LISTType;
                            Data : ADDRESS;
                        VAR Found: BOOLEAN);
  (* Find Data in the list. If Found, go there *)

  VAR
    SavePos: CARDINAL;

  BEGIN
    Found := FALSE;
    IF (List^.N=0) THEN RETURN END;
    (* If the cursor is undefined, start at the top *)
    IF (List^.CurPos=0) THEN
      List^.Cursor := List^.Top^.Next;
      List^.CurPos := 1
    END;
    (* Save current position *)
    SavePos := List^.CurPos;
    (* Loop, until the current position is reached again or Data is found *)
    REPEAT
      (* Check the current element *)
      Found := (Classes[List^.ClassID].Cmp(Data,List^.Cursor^.DataPtr)
                =LISTEqual);
      (* Not found --> search on cyclic *)
      IF (NOT Found) THEN
        IF (List^.CurPos=List^.N) THEN
          List^.Cursor := List^.Top^.Next;
          List^.CurPos := 1
        ELSE
          List^.Cursor := List^.Cursor^.Next;
          INC(List^.CurPos)
        END
      END
    UNTIL (Found) OR (List^.CurPos=SavePos)
  END GotoElement;

  PROCEDURE InsertAfterCursor(VAR List: LISTType;
                                  Data: ADDRESS);
  (* Insert a new node after the current position and update all necessary
     information *)

  VAR
    Help: NodeTypePtr;

  BEGIN
    (* Create a help node for the new data *)
    ALLOCATE(Help,TSIZE(NodeType));
    ALLOCATE(Help^.DataPtr,Classes[List^.ClassID].Bytes);
    (* Copy data *)
    Copy(Data,Help^.DataPtr,Classes[List^.ClassID].Bytes);
    (* List empty --> position is top *)
    IF (List^.N=0) THEN List^.Cursor := List^.Top END;
    (* insert new element behind the current one *)
    Help^.Next := List^.Cursor^.Next;
    Help^.Prev := List^.Cursor;
    List^.Cursor^.Next^.Prev := Help;
    List^.Cursor^.Next := Help;
    (* update the cursor and the organisatoric values *)
    List^.Cursor := Help;
    INC(List^.N);
    INC(List^.CurPos)
  END InsertAfterCursor;

  PROCEDURE InsertBeforeCursor(VAR List: LISTType;
                                   Data: ADDRESS);
  (* Insert a new node before the current position and update all necessary
     information *)

  VAR
    Help: NodeTypePtr;

  BEGIN
    (* Create a help node for the new data *)
    ALLOCATE(Help,TSIZE(NodeType));
    ALLOCATE(Help^.DataPtr,Classes[List^.ClassID].Bytes);
    (* Copy data *)
    Copy(Data,Help^.DataPtr,Classes[List^.ClassID].Bytes);
    (* List empty --> position is top *)
    IF (List^.N=0) THEN
      List^.Cursor := List^.Bottom;
      List^.CurPos := 1
    END;
    (* Insert the new element before the current position *)
    Help^.Next := List^.Cursor;
    Help^.Prev := List^.Cursor^.Prev;
    List^.Cursor^.Prev^.Next := Help;
    List^.Cursor^.Prev := Help;
    (* update the cursor and the organisatoric values *)
    List^.Cursor := Help;
    INC(List^.N)
  END InsertBeforeCursor;

  PROCEDURE MoveCurrentBefore(VAR List: LISTType;
                                  To  : NodeTypePtr);
  (* Move the current element before the element To. The cursor must be
     defined. CursorPos must be set hereafter *)

  BEGIN
    (* Special case: Current is already predecessor of To *)
    IF (List^.Cursor^.Next=To) THEN RETURN END;
    (* Special case: Current is successor of To *)
    IF (List^.Cursor^.Prev=To) THEN
      LISTExchangeCurrentPrevious(List);
      RETURN
    END;
    (* Normal case: The elements are no neighbours *)
    List^.Cursor^.Next^.Prev := List^.Cursor^.Prev;
    List^.Cursor^.Prev^.Next := List^.Cursor^.Next;
    List^.Cursor^.Next := To;
    List^.Cursor^.Prev := To^.Prev;
    To^.Prev^.Next := List^.Cursor;
    To^.Prev := List^.Cursor
  END MoveCurrentBefore;

  PROCEDURE MoveCurrentAfter(VAR List: LISTType;
                                 To  : NodeTypePtr);
  (* Move the current element after the element To. The cursor must be
     defined. CursorPos must be set hereafter *)

  BEGIN
    (* Special case: Current is already successor of To *)
    IF (List^.Cursor^.Prev=To) THEN RETURN END;
    (* Special case: Current is predecessor of To *)
    IF (List^.Cursor^.Next=To) THEN
      LISTExchangeCurrentNext(List);
      RETURN
    END;
    (* Normal case: The elements are no neighbours *)
    List^.Cursor^.Next^.Prev := List^.Cursor^.Prev;
    List^.Cursor^.Prev^.Next := List^.Cursor^.Next;
    List^.Cursor^.Prev := To;
    List^.Cursor^.Next := To^.Next;
    To^.Next^.Prev := List^.Cursor;
    To^.Next := List^.Cursor
  END MoveCurrentAfter;

  (**********************************************************************)
  (*                                                                    *)
  (* 			    Exported Procedures				*)
  (*                                                                    *)
  (**********************************************************************)

  PROCEDURE LISTCreateClass(    CmpProc: LISTCompareProcType;
                                Bytes  : CARDINAL;
                            VAR ClassID: CARDINAL);

  BEGIN
    Classes[ClassN].Cmp := CmpProc;
    Classes[ClassN].Bytes := Bytes;
    ClassID := ClassN;
    INC(ClassN)
  END LISTCreateClass;

  PROCEDURE LISTCreate(VAR List : LISTType;
                           Class: CARDINAL);

  BEGIN
    (* Speicher fuer Liste holen *)
    ALLOCATE(List,TSIZE(ListType));
    WITH List^ DO
      (* Create a dummy top and a dummy bottom node. These nodes will never
	 be deleted (unless the list is killed) and exist just for efficiency
	 reasons *)
      ALLOCATE(Top,TSIZE(NodeType));
      ALLOCATE(Bottom,TSIZE(NodeType));
      Top^.Prev := NIL;
      Bottom^.Next := NIL;
      Top^.Next := Bottom;
      Bottom^.Prev := Top;
      CurPos := 0;
      ClassID := Class;
      N := 0
    END
  END LISTCreate;

  PROCEDURE LISTSetUndefined(VAR List: LISTType);

  BEGIN List := NIL
  END LISTSetUndefined;

  PROCEDURE LISTIsDefined(    List: LISTType): BOOLEAN;

  BEGIN RETURN(List#NIL)
  END LISTIsDefined;

  PROCEDURE LISTLength(    List: LISTType): CARDINAL;

  BEGIN 
    RETURN(List^.N)
  END LISTLength;

  PROCEDURE LISTTop(    List: LISTType): BOOLEAN;

  BEGIN 
    (* Note: LISTTop is also TRUE, if the list is empty (N=0) *)
    RETURN(List^.CurPos<2)
  END LISTTop;

  PROCEDURE LISTBottom(    List: LISTType): BOOLEAN;

  BEGIN 
    (* Note: LISTBottom is also TRUE, if the list is empty (CurPos=N=0) *)
    RETURN(List^.CurPos=List^.N)
  END LISTBottom;

  PROCEDURE LISTCursorPos(    List: LISTType): CARDINAL;

  BEGIN 
    (* Note: CurPos is Zero, if the cursor is undefined *)
    RETURN(List^.CurPos)
  END LISTCursorPos;

  PROCEDURE LISTIsEmpty(    List: LISTType): BOOLEAN;

  BEGIN 
    RETURN(List^.N=0)
  END LISTIsEmpty;

  PROCEDURE LISTClear(VAR List: LISTType);

  VAR
    Help: NodeTypePtr;

  BEGIN
    List^.Cursor := List^.Top^.Next;
    WHILE (List^.Cursor#List^.Bottom) DO
      Help := List^.Cursor^.Next;
      DEALLOCATE(List^.Cursor^.DataPtr,Classes[List^.ClassID].Bytes);
      DEALLOCATE(List^.Cursor,TSIZE(NodeType));
      List^.Cursor := Help
    END;
    (* make cursor undefined *)
    List^.CurPos := 0;
    List^.N := 0;
    (* recreate top/bottom linkage *)
    List^.Top^.Next := List^.Bottom;
    List^.Bottom^.Prev := List^.Top
  END LISTClear;

  PROCEDURE LISTGotoTop(VAR List: LISTType);

  BEGIN
    IF (List^.N=0) THEN RETURN END;
    List^.Cursor := List^.Top^.Next;
    List^.CurPos := 1
  END LISTGotoTop;

  PROCEDURE LISTGotoBottom(VAR List: LISTType);

  BEGIN
    IF (List^.N=0) THEN RETURN END;
    List^.Cursor := List^.Bottom^.Prev;
    List^.CurPos := List^.N
  END LISTGotoBottom;

  PROCEDURE LISTGotoPrevious(VAR List: LISTType);

  BEGIN
    (* make cursor undefined, if top is reached *)
    IF (List^.CurPos<2) THEN 
      List^.CurPos := 0
    ELSE
      List^.Cursor := List^.Cursor^.Prev;
      DEC(List^.CurPos)
    END
  END LISTGotoPrevious;

  PROCEDURE LISTGotoNext(VAR List: LISTType);

  BEGIN
    (* make cursor undefined, if bottom is reached *)
    IF (List^.CurPos=List^.N) THEN
      List^.CurPos := 0
    ELSIF (List^.CurPos<>0) THEN
      List^.Cursor := List^.Cursor^.Next;
      INC(List^.CurPos)
    END
  END LISTGotoNext;

  PROCEDURE LISTGotoElement(VAR List : LISTType;
                            VAR Data : ARRAY OF BYTE;
                            VAR Found: BOOLEAN);

  BEGIN
    GotoElement(List,ADR(Data),Found) 
  END LISTGotoElement;

  PROCEDURE LISTStep(VAR List : LISTType;
                         Steps: INTEGER);

  BEGIN
    (* go forward in Next direction *)
    WHILE (Steps>0) AND (List^.Cursor#List^.Bottom) DO
      List^.Cursor := List^.Cursor^.Next;
      INC(List^.CurPos);
      DEC(Steps)
    END;
    (* go backward in Previous direction *)
    WHILE (Steps<0) AND (List^.Cursor#List^.Top) DO
      List^.Cursor := List^.Cursor^.Prev;
      DEC(List^.CurPos);
      INC(Steps)
    END;
  END LISTStep;

  PROCEDURE LISTCyclicStep(VAR List : LISTType;
                               Steps: INTEGER);

  BEGIN
    (* go forward in Next direction *)
    WHILE (Steps>0) DO
      List^.Cursor := List^.Cursor^.Next;
      (* bottom reached --> goto top *)
      IF (List^.Cursor=List^.Bottom) THEN
        List^.Cursor := List^.Top^.Next;
        List^.CurPos := 1
      ELSE
        INC(List^.CurPos);
      END;
      DEC(Steps)
    END;
    (* go backward in Previous direction *)
    WHILE (Steps<0) DO
      List^.Cursor := List^.Cursor^.Prev;
      (* top reached --> goto bottom *)
      IF (List^.Cursor=List^.Top) THEN
        List^.Cursor := List^.Bottom^.Prev;
        List^.CurPos := List^.N
      ELSE
        DEC(List^.CurPos)
      END;
      INC(Steps)
    END
  END LISTCyclicStep;

  PROCEDURE LISTGotoPos(VAR List: LISTType;
                            Pos : CARDINAL);

  VAR
    Dist1,Dist2,Dist3: INTEGER;
    Abs1,Abs2,Abs3   : CARDINAL;
    Min23,Min13      : CARDINAL;

  BEGIN
    (* cursor undefined --> goto bottom *)
    IF (List^.CurPos=0) THEN
      List^.Cursor := List^.Bottom^.Prev;
      List^.CurPos := List^.N
    END;
    (* Calculate the distances from Top, Bottom, and current position with and
       without sign *)
    Dist1 := INTEGER(Pos)-INTEGER(1);
    IF (Dist1>0) THEN Abs1 := CARDINAL(Dist1) ELSE Abs1 := CARDINAL(-Dist1) END;
    Dist2 := INTEGER(Pos)-INTEGER(List^.CurPos);
    IF (Dist2>0) THEN Abs2 := CARDINAL(Dist2) ELSE Abs2 := CARDINAL(-Dist2) END;
    Dist3 := INTEGER(Pos)-INTEGER(List^.N);
    IF (Dist3>0) THEN Abs3 := CARDINAL(Dist3) ELSE Abs3 := CARDINAL(-Dist3) END;
    (* Calculate the minimum distance *)
    IF (Abs2<Abs3) THEN Min23 := Abs2 ELSE Min23 := Abs3 END;
    IF (Abs1<Abs3) THEN Min13 := Abs1 ELSE Min13 := Abs3 END;
    (* Check which position is nearest *)
    IF (Abs1<=Min23) THEN
      (* top is nearest *)
      List^.Cursor := List^.Top^.Next;
      List^.CurPos := 1;
      (* go forward to the position *)
      LISTStep(List,Dist1)
    ELSIF (Abs2<=Min13) THEN
      (* Cursor is nearest --> go from here *)
      LISTStep(List,Dist2)
    ELSE
      (* bottom is nearest *)
      List^.Cursor := List^.Bottom^.Prev;
      List^.CurPos := List^.N;
      (* go backward to the position *)
      LISTStep(List,Dist3)
    END
  END LISTGotoPos;

  PROCEDURE LISTKill(VAR List: LISTType);

  BEGIN
    (* Delete the list contents *)
    LISTClear(List);
    (* Delete top and bottom *)
    DEALLOCATE(List^.Top,TSIZE(NodeType));
    DEALLOCATE(List^.Bottom,TSIZE(NodeType));
    (* Delete the list header *)
    DEALLOCATE(List,TSIZE(ListType));
    List := NIL
  END LISTKill;

  PROCEDURE LISTRemoveElement(VAR List: LISTType;
                              VAR Data: ARRAY OF BYTE);

  VAR
    Found: BOOLEAN;

  BEGIN
    GotoElement(List,ADR(Data),Found); 
    IF (Found) THEN DeleteCurrent(List) END
  END LISTRemoveElement;

  PROCEDURE LISTDeleteElement(VAR List : LISTType;
                              VAR Data : ARRAY OF BYTE;
                              VAR Found: BOOLEAN);

  BEGIN
    GotoElement(List,ADR(Data),Found); 
    IF (Found) THEN DeleteCurrent(List) END
  END LISTDeleteElement;

  PROCEDURE LISTRemoveCurrent(VAR List: LISTType);

  BEGIN
    DeleteCurrent(List)
  END LISTRemoveCurrent;

  PROCEDURE LISTExtractCurrent(VAR List : LISTType;
                               VAR Data : ARRAY OF BYTE;
                               VAR Found: BOOLEAN);

  BEGIN
    IF (List^.CurPos=0) THEN Found := FALSE; RETURN END;
    Found := TRUE;
    Copy(List^.Cursor^.DataPtr,ADR(Data),Classes[List^.ClassID].Bytes); 
    DeleteCurrent(List)
  END LISTExtractCurrent;

  PROCEDURE LISTGetCurrent(VAR List : LISTType;
                           VAR Data : ARRAY OF BYTE;
                           VAR Found: BOOLEAN);

  BEGIN
    IF (List^.CurPos=0) THEN Found := FALSE; RETURN END;
    Found := TRUE;
    Copy(List^.Cursor^.DataPtr,ADR(Data),Classes[List^.ClassID].Bytes) 
  END LISTGetCurrent;

  PROCEDURE LISTReplaceCurrent(VAR List: LISTType;
                               VAR Data: ARRAY OF BYTE);

  BEGIN
    Copy(ADR(Data),List^.Cursor^.DataPtr,Classes[List^.ClassID].Bytes) 
  END LISTReplaceCurrent;

  PROCEDURE LISTInsertTop(VAR List: LISTType;
                          VAR Data: ARRAY OF BYTE);

  BEGIN
    List^.CurPos := 1;
    List^.Cursor := List^.Top^.Next;
    InsertBeforeCursor(List,ADR(Data)) 
  END LISTInsertTop;

  PROCEDURE LISTInsertBottom(VAR List: LISTType;
                             VAR Data: ARRAY OF BYTE);

  BEGIN
    List^.CurPos := List^.N;
    List^.Cursor := List^.Bottom^.Prev;
    InsertAfterCursor(List,ADR(Data)) 
  END LISTInsertBottom;

  PROCEDURE LISTInsertBeforeCursor(VAR List: LISTType;
                                   VAR Data: ARRAY OF BYTE);

  BEGIN
    InsertBeforeCursor(List,ADR(Data)) 
  END LISTInsertBeforeCursor;

  PROCEDURE LISTInsertAfterCursor(VAR List: LISTType;
                                  VAR Data: ARRAY OF BYTE);

  BEGIN
    InsertAfterCursor(List,ADR(Data)) 
  END LISTInsertAfterCursor;

  PROCEDURE LISTExchangeCurrentPrevious(VAR List: LISTType);
                              
  BEGIN
    (* change linkage in Next direction *)
    List^.Cursor^.Prev^.Prev^.Next := List^.Cursor;
    List^.Cursor^.Prev^.Next := List^.Cursor^.Next;
    List^.Cursor^.Next := List^.Cursor^.Prev;
    (* change linkage in Previous direction *)
    List^.Cursor^.Prev := List^.Cursor^.Prev^.Prev;
    List^.Cursor^.Next^.Prev := List^.Cursor;
    List^.Cursor^.Next^.Next^.Prev := List^.Cursor^.Next;
    (* update position *)
    DEC(List^.CurPos)
  END LISTExchangeCurrentPrevious;

  PROCEDURE LISTExchangeCurrentNext(VAR List: LISTType);

  BEGIN
    (* change linkage in Next direction *)
    List^.Cursor^.Next^.Next^.Prev := List^.Cursor;
    List^.Cursor^.Next^.Prev := List^.Cursor^.Prev;
    List^.Cursor^.Prev := List^.Cursor^.Next;
    (* change linkage in Previous direction *)
    List^.Cursor^.Next := List^.Cursor^.Next^.Next;
    List^.Cursor^.Prev^.Next := List^.Cursor;
    List^.Cursor^.Prev^.Prev^.Next := List^.Cursor^.Prev;
    (* update position *)
    INC(List^.CurPos)
  END LISTExchangeCurrentNext;

  PROCEDURE LISTExchangePos(VAR List: LISTType;
                                Pos1: CARDINAL;
                                Pos2: CARDINAL);

  VAR
    Help : NodeTypePtr;
    Help2: NodeTypePtr;

  BEGIN
    (* if Pos1=Pos2, return without action *)
    IF (Pos1=Pos2) THEN RETURN END;
    (* goto Pos1 and store pointer *)
    LISTGotoPos(List,Pos1);
    Help := List^.Cursor;
    (* goto Pos2 *)
    LISTGotoPos(List,Pos2);
    (* special cases: the nodes are neighbours *)
    IF (Help^.Next=List^.Cursor) THEN 
      LISTExchangeCurrentPrevious(List); RETURN
    ELSIF (Help=List^.Cursor^.Next) THEN 
      LISTExchangeCurrentNext(List); RETURN 
    END;
    (* normal case: the nodes are no neighbours. Then change the linkage *)
    List^.Cursor^.Next^.Prev := Help; Help^.Prev^.Next := List^.Cursor;
    List^.Cursor^.Prev^.Next := Help; Help^.Next^.Prev := List^.Cursor;
    Help2 := Help^.Next;
    Help^.Next := List^.Cursor^.Next; List^.Cursor^.Next := Help2;
    Help2 := Help^.Prev;
    Help^.Prev := List^.Cursor^.Prev; List^.Cursor^.Prev := Help2;
    (* update position *)
    List^.CurPos := Pos1
  END LISTExchangePos;

  PROCEDURE LISTMoveCurrentToTop(VAR List: LISTType);

  BEGIN
    MoveCurrentBefore(List,List^.Top^.Next);
    List^.CurPos := 1
  END LISTMoveCurrentToTop;

  PROCEDURE LISTMoveCurrentToBottom(VAR List: LISTType);

  BEGIN
    MoveCurrentAfter(List,List^.Bottom^.Prev);
    List^.CurPos := List^.N
  END LISTMoveCurrentToBottom;

  PROCEDURE LISTMoveElementBeforePos(VAR List: LISTType;
                                         From: CARDINAL;
                                         To  : CARDINAL);

  VAR
    ToPtr: NodeTypePtr;

  BEGIN
    (* goto To and store pointer *)
    LISTGotoPos(List,To);
    ToPtr := List^.Cursor;
    (* goto From *)
    LISTGotoPos(List,From);
    MoveCurrentBefore(List,ToPtr);
    IF (From<To) THEN List^.CurPos := To-1 ELSE List^.CurPos := To END
  END LISTMoveElementBeforePos;

  PROCEDURE LISTMoveElementAfterPos(VAR List: LISTType;
                                        From: CARDINAL;
                                        To  : CARDINAL);

  VAR
    ToPtr: NodeTypePtr;

  BEGIN
    (* goto To and store pointer *)
    LISTGotoPos(List,To);
    ToPtr := List^.Cursor;
    (* goto From *)
    LISTGotoPos(List,From);
    MoveCurrentAfter(List,ToPtr);
    IF (From<To) THEN List^.CurPos := To ELSE List^.CurPos := To-1 END
  END LISTMoveElementAfterPos;

  PROCEDURE LISTCopyCurrentToTop(VAR List: LISTType);

  VAR
    Adr: ADDRESS;

  BEGIN
    Adr := List^.Cursor^.DataPtr;
    List^.CurPos := 1;
    List^.Cursor := List^.Top^.Next;
    InsertBeforeCursor(List,Adr)
  END LISTCopyCurrentToTop;

  PROCEDURE LISTCopyCurrentToBottom(VAR List: LISTType);

  VAR
    Adr: ADDRESS;

  BEGIN
    Adr := List^.Cursor^.DataPtr;
    List^.CurPos := List^.N;
    List^.Cursor := List^.Bottom^.Prev;
    InsertAfterCursor(List,Adr)
  END LISTCopyCurrentToBottom;

  PROCEDURE LISTCopyElementBeforePos(VAR List: LISTType;
                                         From: CARDINAL;
                                         To  : CARDINAL);

  VAR
    Adr: ADDRESS;

  BEGIN
    LISTGotoPos(List,From);
    Adr := List^.Cursor^.DataPtr;
    LISTGotoPos(List,To);
    InsertBeforeCursor(List,Adr)
  END LISTCopyElementBeforePos;

  PROCEDURE LISTCopyElementAfterPos(VAR List: LISTType;
                                        From: CARDINAL;
                                        To  : CARDINAL);

  VAR
    Adr: ADDRESS;

  BEGIN
    LISTGotoPos(List,From);
    Adr := List^.Cursor^.DataPtr;
    LISTGotoPos(List,To);
    InsertAfterCursor(List,Adr)
  END LISTCopyElementAfterPos;

  PROCEDURE LISTCopy(    Source: LISTType;
                     VAR Target: LISTType);

  VAR
    Help   : NodeTypePtr;
    NewNode: NodeTypePtr;

  BEGIN
    (* create target list *)
    LISTCreate(Target,Source^.ClassID);
    (* loop over the source list *)
    Help := Source^.Top^.Next;
    WHILE (Help#Source^.Bottom) DO
      (* insert a new node at the end of the target list *)
      ALLOCATE(NewNode,TSIZE(NodeType));
      ALLOCATE(NewNode^.DataPtr,Classes[Source^.ClassID].Bytes);
      Copy(Help^.DataPtr,NewNode^.DataPtr,Classes[Source^.ClassID].Bytes);
      NewNode^.Prev := Target^.Bottom^.Prev;
      NewNode^.Next := Target^.Bottom;
      Target^.Bottom^.Prev^.Next := NewNode;
      Target^.Bottom^.Prev := NewNode;
      (* if the source cursor is reached, set the target cursor to the
         corresponding element *)
      IF (Help=Source^.Cursor) THEN Target^.Cursor := NewNode END;
      Help := Help^.Next
    END;
    (* update position *)
    Target^.CurPos := Source^.CurPos;
    Target^.N := Source^.N
  END LISTCopy;

  PROCEDURE LISTInsertListBeforeCursor(VAR FirstList : LISTType;
                                       VAR SecondList: LISTType);

  BEGIN
    (* Special case: second list empty *)
    IF (SecondList^.N=0) THEN
      LISTKill(SecondList);
      RETURN
    END;
    (* insert the second list before the cursor of the first one *)
    FirstList^.Cursor^.Prev^.Next := SecondList^.Top^.Next;
    SecondList^.Top^.Next^.Prev := FirstList^.Cursor^.Prev;
    FirstList^.Cursor^.Prev := SecondList^.Bottom^.Prev;
    SecondList^.Bottom^.Prev^.Next := FirstList^.Cursor;
    (* update position *)
    INC(FirstList^.N,SecondList^.N);
    INC(FirstList^.CurPos,SecondList^.N);
    (* Delete top and bottom *)
    DEALLOCATE(SecondList^.Top,TSIZE(NodeType));
    DEALLOCATE(SecondList^.Bottom,TSIZE(NodeType));
    (* Delete second list *)
    DEALLOCATE(SecondList,TSIZE(LISTType));
    SecondList := NIL
  END LISTInsertListBeforeCursor;

  PROCEDURE LISTInsertListAfterCursor(VAR FirstList : LISTType;
                                      VAR SecondList: LISTType);

  BEGIN
    (* Special case: second list empty *)
    IF (SecondList^.N=0) THEN
      LISTKill(SecondList);
      RETURN
    END;
    (* Insert the second list after the current element of the first one.
       The cursor remains unchanged *)
    FirstList^.Cursor^.Next^.Prev := SecondList^.Bottom^.Prev;
    SecondList^.Bottom^.Prev^.Next := FirstList^.Cursor^.Next;
    FirstList^.Cursor^.Next := SecondList^.Top^.Next;
    SecondList^.Top^.Next^.Prev := FirstList^.Cursor;
    INC(FirstList^.N,SecondList^.N);
    (* Delete top and bottom *)
    DEALLOCATE(SecondList^.Top,TSIZE(NodeType));
    DEALLOCATE(SecondList^.Bottom,TSIZE(NodeType));
    (* delete the second list *)
    DEALLOCATE(SecondList,TSIZE(LISTType));
    SecondList := NIL
  END LISTInsertListAfterCursor;

  PROCEDURE LISTSplitListBeforeCursor(VAR Source: LISTType;
                                      VAR Tail  : LISTType);

  VAR
    Help: NodeTypePtr;

  BEGIN
    (* redirect pointers *)
    Help := Tail^.Bottom^.Prev;
    Source^.Cursor^.Prev^.Next := Source^.Bottom;
    Tail^.Bottom^.Prev^.Next := Source^.Cursor;
    Source^.Bottom^.Prev^.Next := Tail^.Bottom;
    Tail^.Bottom^.Prev := Source^.Bottom^.Prev;
    Source^.Bottom^.Prev := Source^.Cursor^.Prev;
    Source^.Cursor^.Prev := Help;
    (* update values *)
    Tail^.CurPos := Tail^.N+1;
    INC(Tail^.N,Source^.N+1-Source^.CurPos);
    Source^.N := Source^.CurPos-1;
    Source^.CurPos := Source^.N;
    Tail^.Cursor := Source^.Cursor;
    Source^.Cursor := Source^.Bottom^.Prev
  END LISTSplitListBeforeCursor;

  PROCEDURE LISTSplitListAfterCursor(VAR Source: LISTType;
                                     VAR Tail  : LISTType);

  BEGIN
    IF (Source^.CurPos=Source^.N) THEN RETURN END;
    (* redirect pointers *)
    Source^.Bottom^.Prev^.Next := Tail^.Bottom;
    Tail^.Bottom^.Prev^.Next := Source^.Cursor^.Next;
    Source^.Cursor^.Next^.Prev := Tail^.Bottom^.Prev;
    Tail^.Bottom^.Prev := Source^.Bottom^.Prev;
    Source^.Bottom^.Prev := Source^.Cursor;
    Source^.Cursor^.Next := Source^.Bottom;
    (* update values *)
    Tail^.CurPos := Tail^.N+1;
    INC(Tail^.N,Source^.N-Source^.CurPos);
    Source^.N := Source^.CurPos;
    Source^.CurPos := Source^.N;
    Tail^.Cursor := Source^.Cursor;
    Source^.Cursor := Source^.Bottom^.Prev
  END LISTSplitListAfterCursor;

  PROCEDURE LISTRemoveRange(VAR List: LISTType;
                                Pos1: CARDINAL;
                                Pos2: CARDINAL);

  VAR
    HelpPtr : NodeTypePtr;
    HelpPtr2: NodeTypePtr;
    HelpPos : CARDINAL;
  
  BEGIN
    (* Pos1=Pos2 --> delete element *)
    IF (Pos1=Pos2) THEN
      LISTGotoPos(List,Pos1);
      LISTRemoveCurrent(List);
      RETURN
    END;
    (* Sort so that Pos1<Pos2 *)
    IF (Pos1>Pos2) THEN HelpPos := Pos1; Pos1 := Pos2; Pos2 := HelpPos END;
    (* Goto Pos1 and store pointer *)
    LISTGotoPos(List,Pos1);
    HelpPtr := List^.Cursor;
    LISTGotoPos(List,Pos2);
    (* Link out the range from HelpPtr to List^.Cursor *)
    HelpPtr^.Prev^.Next := List^.Cursor^.Next;
    List^.Cursor^.Next^.Prev := HelpPtr^.Prev;
    List^.Cursor := List^.Cursor^.Next;
    (* free memory *)
    WHILE (HelpPtr#List^.Cursor) DO
      HelpPtr2 := HelpPtr^.Next;
      DEALLOCATE(HelpPtr^.DataPtr,Classes[List^.ClassID].Bytes);
      DEALLOCATE(HelpPtr,TSIZE(NodeType));
      HelpPtr := HelpPtr2
    END;
    List^.CurPos := Pos1;
    (* Determine new cursor position (sucessor or predecessor, if there is
       no sucessor) *)
    IF (List^.Cursor=List^.Bottom) THEN
      List^.Cursor := List^.Cursor^.Prev;
      (* cursor becomes undefined if Pos1=1, Pos2=N *)
      DEC(List^.CurPos)
    END;
    DEC(List^.N,Pos2-Pos1+1)
  END LISTRemoveRange;

  PROCEDURE LISTLoop(VAR List: LISTType);
  
  BEGIN
    IF (List^.N=0) THEN RETURN END;
    List^.CurPos := 1;
    List^.Cursor := List^.Top^.Next
  END LISTLoop;
  
  PROCEDURE LISTGet(VAR List : LISTType;
                    VAR Data : ARRAY OF BYTE;
                    VAR Found: BOOLEAN);

  BEGIN
    IF (List^.CurPos=0) THEN Found := FALSE; RETURN END;
    Found := TRUE;
    Copy(List^.Cursor^.DataPtr,ADR(Data),Classes[List^.ClassID].Bytes); 
    List^.Cursor := List^.Cursor^.Next;
    INC(List^.CurPos);
    (* set cursor to undefined if the end is reached *)
    IF (List^.CurPos>List^.N) THEN List^.CurPos := 0 END
  END LISTGet;    

  PROCEDURE LISTInvert(VAR List: LISTType);

  VAR
    Help: NodeTypePtr;
    Ptr : NodeTypePtr;
  
  BEGIN
    (* Swap all pointers from Next to Previous and vice versa *)
    Help := List^.Top;
    WHILE (Help#NIL) DO
      Ptr := Help^.Next; Help^.Next := Help^.Prev; Help^.Prev := Ptr;
      Help := Help^.Prev
    END;
    (* swap top and bottom *)
    Ptr := List^.Top; List^.Top := List^.Bottom; List^.Bottom := Ptr;
    (* Update position *)
    IF (List^.CurPos#0) THEN List^.CurPos := List^.N+1-List^.CurPos END
  END LISTInvert;

  PROCEDURE LISTPack(VAR List: LISTType);

  VAR
    Vgl: ADDRESS; 
    Help: NodeTypePtr;
    
  BEGIN
    (* exit if there are less than 2 elements in the list *)
    IF (List^.N<2) THEN RETURN END;
    (* compare with the first element *)
    Vgl := List^.Top^.Next^.DataPtr;
    (* start a loop over the list *)
    List^.Cursor := List^.Top^.Next^.Next;
    WHILE (List^.Cursor#List^.Bottom) DO
      (* Note: Help will have the correct value no matter whether Cursor will
	 be deleted or not *)
      Help := List^.Cursor^.Next;
      (* Delete the current element if its equal to Vgl *)
      IF (Classes[List^.ClassID].Cmp(Vgl,List^.Cursor^.DataPtr)=LISTEqual) THEN
        DeleteCurrent(List)
      ELSE
        Vgl := List^.Cursor^.DataPtr
      END;
      List^.Cursor := Help
    END;
    (* set the cursor to Top. Note: The list cannot become empty! *)
    List^.Cursor := List^.Top^.Next;
    List^.CurPos := 1    
  END LISTPack;

  PROCEDURE LISTSort(VAR List     : LISTType;
                         Ascending: BOOLEAN);
  (* Merge sort algorithm for lists - taken from G.H. Gonnet, "Handbook of
     Algorithms and Data Structures", Addison-Wesley *)

  VAR
    Last  : NodeTypePtr;
    Result: LISTCompareResultType;

    PROCEDURE TailIns(    Rec  : NodeTypePtr;
                      VAR First: NodeTypePtr;
                      VAR Last : NodeTypePtr);

    BEGIN
      IF (First=NIL) THEN First := Rec ELSE Last^.Next := Rec END;
      Last := Rec
    END TailIns;

    PROCEDURE MergeLists(    a: NodeTypePtr;
                             b: NodeTypePtr): NodeTypePtr;
    
    VAR
      First: NodeTypePtr;
      Last : NodeTypePtr;
      Help : NodeTypePtr;

    BEGIN
      First := NIL;
      WHILE (b#NIL) DO
        IF (a=NIL) THEN
          a := b; b := NIL
        ELSE
          IF (Classes[List^.ClassID].Cmp(b^.DataPtr,a^.DataPtr)=Result)
          THEN
            Help := a; a := a^.Next
          ELSE
            Help := b; b := b^.Next
          END;
          Help^.Next := NIL;
          TailIns(Help,First,Last)
        END
      END;
      TailIns(a,First,Last);
      RETURN(First)
    END MergeLists;

    PROCEDURE MergeSort(VAR Root: NodeTypePtr;
                            N   : CARDINAL): NodeTypePtr;
    
    VAR
      Help: NodeTypePtr;
      a,b : NodeTypePtr;

    BEGIN
      IF (Root=NIL) THEN
        RETURN(NIL)
      ELSIF (N>1) THEN
        a := MergeSort(Root,N DIV 2);
        b := MergeSort(Root,(N+1) DIV 2);
        RETURN(MergeLists(a,b))
      ELSE
        Help := Root;
        Root := Root^.Next;
        Help^.Next := NIL;
        RETURN(Help)
      END
    END MergeSort;

  BEGIN
    IF (List^.N<2) THEN RETURN END;
    (* init compare result depending on the sorting order *)
    IF (Ascending) THEN Result := LISTGreater ELSE Result := LISTLess END;
    (* do the sort *)
    List^.Top^.Next := MergeSort(List^.Top^.Next,List^.N);
    (* relink the nodes in Previous direction *)
    Last := List^.Top;
    List^.Cursor := List^.Top^.Next;
    WHILE (List^.Cursor#NIL) DO
      List^.Cursor^.Prev := Last;
      Last := List^.Cursor;
      List^.Cursor := List^.Cursor^.Next
    END;
    Last^.Next := List^.Bottom;
    List^.Bottom^.Prev := Last;
    (* set cursor to top *)
    List^.CurPos := 1;
    List^.Cursor := List^.Top^.Next
  END LISTSort;

  PROCEDURE LISTInsertSorted(VAR List     : LISTType;
                             VAR Data     : ARRAY OF BYTE;
                                 Ascending: BOOLEAN);

  VAR
    Result: LISTCompareResultType;
    
  BEGIN
    IF (Ascending) THEN Result := LISTLess ELSE Result := LISTGreater END;
    List^.Cursor := List^.Top^.Next;
    List^.CurPos := 1;
    (* loop until the insertion place is found *)
    LOOP
      IF (List^.Cursor=List^.Bottom) THEN EXIT END;
      IF (Classes[List^.ClassID].Cmp(List^.Cursor^.DataPtr,ADR(Data))#Result) 
      THEN
        InsertBeforeCursor(List,ADR(Data));
        RETURN
      END;
      List^.Cursor := List^.Cursor^.Next;
      INC(List^.CurPos)
    END;
    (* Special case: insertion at the end *)
    List^.CurPos := List^.N;
    List^.Cursor := List^.Bottom^.Prev;
    InsertAfterCursor(List,ADR(Data)) 
  END LISTInsertSorted;

  (**********************************************************************)
  (*                                                                    *)
  (*                           Testprozedur                             *)
  (*                                                                    *)
  (**********************************************************************)

  PROCEDURE LISTCheckConsistency(    List: LISTType);

  VAR
    Help: NodeTypePtr;
    i   : CARDINAL;

  BEGIN
    IF (List=NIL) THEN
      HALT
    END;
    (* Test 1: Go from Top to Bottom and check the cursor, if defined *)
    Help := List^.Top;
    FOR i:=1 TO 1+List^.N DO
      IF (List^.CurPos#0) THEN
        IF (Help=List^.Cursor) THEN
          IF (i-1#List^.CurPos) THEN
            HALT
          END
        END
      END;
      IF (Help=NIL) THEN
        HALT
      END;
      Help := Help^.Next
    END;
    IF (Help#List^.Bottom) THEN
      HALT
    END;
    (* Test 2: Go from Bottom to Top *)
    FOR i:=1 TO 1+List^.N DO
      IF (Help=NIL) THEN
        HALT
      END;
      Help := Help^.Prev
    END;
    IF (Help#List^.Top) THEN
      HALT
    END
  END LISTCheckConsistency;

BEGIN ClassN := 0
END LISTs.
